__license__ = """
Sonata, an elegant GTK+ client for the Music Player Daemon

Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
Copyright 2009-2014 Jonathan Ballet <jon@multani.info>

This file is part of Sonata.

Sonata is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Sonata is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Sonata.  If not, see <http://www.gnu.org/licenses/>.
"""
